;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "helm-projectile" "20260724.27"
  "Helm integration for Projectile."
  '((emacs      "28.1")
    (helm       "3.0")
    (projectile "3.3.0"))
  :url "https://github.com/bbatsov/helm-projectile"
  :commit "4dae1d072cc2650749846cfcab1f60686471cc45"
  :revdesc "4dae1d072cc2"
  :keywords '("project" "convenience"))
